#! /usr/bin/env python
# -*- encoding=utf8 -*-

"""
A collection of utility functions for managing a decanter project.
Normally you would call this from the command line, such as

    decanter create myproject

which creates a base project called `myproject` to start working in.
"""

import os
import argparse
import errno


def create_project(args):
    print "Creating your project..."
    project_name = args.name
    create_dirs = [
        project_name,
        os.path.join(project_name, 'app/bundles'),
        os.path.join(project_name, 'app/config')
    ]
    for d in create_dirs:
        try:
            os.makedirs(d)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

    # creating files is :TODO:
    create_files = [
    ]

    print "Directory structure created!"


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog='decanter-admin',
        usage='%(prog)s command [options]',
        description="Manage your Decanter project")
    subparsers = parser.add_subparsers()

    # create a sub-parser for the "create" command
    parser_create = subparsers.add_parser(
        'create', help='create a new Decanter project')
    parser_create.add_argument('name', type=str, help='name parameter')
    parser_create.set_defaults(func=create_project)

    args = parser.parse_args()
    args.func(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from gettext import gettext as _
from functools import wraps
import bottle
from bottle import request, redirect
import plugin as lib_plugin
from errors import ValidationError
from . import jsonvalidation


def route(path=None, method='GET', func=None, name=None, apply=None, skip=None, **config):
    def decorator(callback):
        plugins = []

        def apply_plugin(plugin):
            cls = ''.join([plugin.capitalize(), 'Plugin'])
            cls = getattr(lib_plugin, cls)
            plugins.append(cls())

        if isinstance(apply, str):
            apply_plugin(apply)
        elif isinstance(apply, list):
            for plugin in apply:
                apply_plugin(plugin)

        rpath = path
        if rpath != '/':
            rpath = rpath.rstrip('/')

        callback = bottle.route(
            path=rpath, method=method, callback=func, name=name,
            apply=plugins, skip=skip, **config)(callback)

        @wraps(callback)
        def wrapper(*args, **kwargs):
            return callback(*args, **kwargs)
        return wrapper
    return decorator


def get(path=None, **kwargs):
    return route(path=path, method='GET', **kwargs)


def post(path=None, **kwargs):
    return route(path=path, method='POST', **kwargs)


def put(path=None, **kwargs):
    return route(path=path, method='PUT', **kwargs)


def delete(path=None, **kwargs):
    return route(path=path, method='DELETE', **kwargs)


def validate_schema(schema, **kwargs):
    """
    Validate input according to some JSON-schema file,
    and return an error object if there is a problem.
    """
    def decorator(callback):
        def convert_to_dict(bottle_form):
            """
            Function for converting a bottle multiple non-unique keys
            dictionary to a normal python dictionary with lists automatically
            created where appropriate, according to the JSON schema.
            """
            d = {}
            for key, value in bottle_form.iterallitems():
                if key not in d:
                    d[key] = value
                    if key in schema.get('properties') and not isinstance(value, list):
                        if 'array' == schema['properties'][key].get('type') or 'array' in schema['properties'][key].get('type'):
                            d[key] = [value]
                else:
                    if isinstance(d[key], list):
                        d[key] += [value]
                    else:
                        d[key] = [d.get(key)] + [value]
            return d

        @wraps(callback)
        def wrapper(*args, **kwargs):
            if request.headers.get("Content-Type") in ("application/json", "application/javascript", "text/javascript"):
                data = request.json
                instance = data
            else:
                data = request.forms
                data.update(request.query)
                instance = convert_to_dict(data)
            request.cleaned_data = instance
            errors = jsonvalidation.get_error_dictionary(
                schema=schema, instance=instance)
            if errors:
                raise ValidationError(
                    message=_("There were errors validating your request."),
                    fields=errors)
            return callback(*args, **kwargs)
        return wrapper
    return decorator


def login_required(login_url=None, login_id='user_id'):
    '''
    Checks if 'login_id' attribute is present in the session,
    and redirects to 'login_url' if it's not.
    '''
    def decorator(callback):
        @wraps(callback)
        def wrapper(*args, **kwargs):
            session = request.environ['express.session']
            if login_id not in session:
                if login_url:
                    redirect(login_url)
                else:
                    raise Exception('Login Required')
                return {}
            return callback(*args, **kwargs)
        return wrapper
    return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import logging
import logging.handlers
from config import Config


class Log(object):
    __state = {}
    """
    simple logging.Logger wrapper
    """
    def __init__(self, filepath=None):
        """ Borg pattern """
        self.__dict__ = self.__state
        if '_log' not in self.__dict__:
            config = Config()
            if filepath:
                self.filepath = filepath
            elif 'filepath' in config.logger:
                self.filepath = config.logger['filepath']
            else:
                self.filepath = '/var/log/decanter.log'

            try:
                handler = logging.handlers.TimedRotatingFileHandler(
                    self.filepath, 'midnight', 1, 10, 'UTF-8')
            except IOError:
                # check if log directory exists
                parent_path = parent_path = os.path.abspath(
                    os.path.dirname(self.filepath))
                if not os.path.exists(parent_path):
                    # try to create the logs directory
                    os.makedirs(parent_path, mode=0o755)
                # and try again
                handler = logging.handlers.TimedRotatingFileHandler(
                    self.filepath, 'midnight', 1, 10, 'UTF-8')

            formatter = logging.Formatter('%(levelname)s - %(asctime)s - ' +
                                          'File:%(pathname)s - Line:' +
                                          '%(lineno)d - Func:' +
                                          '%(funcName)s\n%(message)s')

            handler.setLevel(config.logger['level'])
            handler.setFormatter(formatter)
            self._log = logging.getLogger(__name__)
            self._log.setLevel(config.logger['level'])
            self._log.addHandler(handler)

    @classmethod
    def get_instance(self):
        return self()

    def critical(self, msg, *args, **kwargs):
        self._log.critical(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self._log.debug(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self._log.error(msg, *args, **kwargs)

    def exception(self, msg, *args):
        self._log.exception(msg, *args)

    def info(self, msg, *args, **kwargs):
        self._log.info(msg, *args, **kwargs)

    def __getattr__(self, name):
        return getattr(self._log, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import bottle
from lib.config import Config


@bottle.error(500)
def error_500(error):
    config = Config()
    accept = bottle.request.headers.get('accept', False)
    is_json = accept and 'json' in accept
    if config.debug and is_json:
        return json.dumps({'status': 500, 'error': error.traceback})
    elif config.debug:
        return bottle.template(bottle.ERROR_PAGE_TEMPLATE, {'e': error})
    elif is_json:
        return json.dumps({'status': 500, 'error': error.body})
    return bottle.template(bottle.ERROR_PAGE_TEMPLATE, {'e': error})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import redis
from config import Config


class Redis(object):

    """ Borg Pattern
    one state many objects useful in this case as we
    want one pool but many objects
    """
    __state = {}

    def __init__(self):
        self.__dict__ = self.__state
        if '_pool' not in self.__dict__:
            config = Config()

            # set defaults
            db = 0
            host = 'localhost'
            port = 6379
            max_connections = 10

            if 'redis' in config:
                db = config.redis.get('db', db)
                host = config.redis.get('host', host)
                port = config.redis.get('port', port)
                max_connections = config.redis.get('max_poolsize', max_connections)

            # we only need one pool per app
            self._pool = redis.ConnectionPool(db=db, host=host, port=port,
                                              max_connections=max_connections)

        # each object required its own connection so that we can close
        # it after each request without affecting other requests
        #
        # so this line is NOT within the above if statement
        self._redis = redis.StrictRedis(connection_pool=self._pool)

    def __getattr__(self, name):
        return getattr(self._redis, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes for loading views and JSON schemas, etc
"""

import os
import codecs
from singleton import Singleton


class BaseLoader(Singleton):

    def __init__(self, config=None):
        """
        Uses the Singleton pattern, so that we can have subclasses
        maintain their own state.
        """
        if getattr(self, '__init', False):
            self.__data = self._load_state(config)

    def __getattr__(self, name):
        return self.__data.get(name, None)

    def __getitem__(self, name):
        return self.__data.get(name, None)

    def __contains__(self, name):
        return name in self.__data

    def _load_state(self, config):
        """
        Method to load the store (for subclasses to override).
        """
        raise NotImplemented()


class JSONLoader(BaseLoader):

    def _load_state(self, config):
        """
        Find all JSON-schema files in the directory structure
        and save them in this state storage.
        """
        data = {}
        path = os.path.join(config.apppath, 'bundles')

        # find all 'validations' folders:
        folders = {}
        for root, dirs, files in os.walk(path):
            validations = filter(lambda f: f == 'validations', dirs)
            folders.update(dict((f, []) for f in map(
                lambda f: os.path.join(root, f), validations)))

        # list all files in these folders
        for k, folder in folders.items():
            for root, dirs, files in os.walk(k):
                files = filter(lambda f: f.endswith('.json'), files)
                folder += map(lambda f: os.path.join(
                    os.path.relpath(root, k), f), files)

        data['directories'] = folders

        return data

    def load_template(self, path):
        rel_path = path
        for full_path, directory in self.directories.items():
            if rel_path in directory:
                return codecs.open(os.path.join(full_path, path), 'r', 'utf-8').read()
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import pwd
import grp

debug = True
test = False

# user name of the user who run decanter
user = pwd.getpwuid(os.getuid()).pw_name

# group name of the user who run decanter
group = grp.getgrgid(os.getegid()).gr_name

# if you run '/path/to/decanter.py', it becomes like '/path/to'
decanterpath = os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.realpath(__file__))))

# files generated by decanter will be placed here
genpath = os.path.join(decanterpath, 'gen')

# pid file
pidfile = os.path.join(genpath, 'run', 'decanter_{0}.py')

# @see: http://bottlepy.org/docs/0.11/api.html#bottle.BaseRequest.MEMFILE_MAX
# Maximum size of memory buffer for body in bytes.
# 20 megabytes
memfile_max = 20971520

# logging
logger = {
    # log directory path, first {0} is the port number and second {1] is the
    # date
    'filepath': os.path.join(genpath, 'log', 'decanter_{0}-{1}.log'),
    # DEBUG, INFO, WARNING, ERROR, FATAL
    'level': 'DEBUG'
}

# the application directory
apppath = os.path.join(decanterpath, 'app')

# default routes
default = {
    'bundle': 'home',
    'controller': 'index'
}

# list of plugins names to install by default
plugins = ['jinja2']

# redis config settings
redis = {
    'db': 0,
    'host': 'localhost',
    'port': 6379,
    'max_poolsize': 5,
}

# session settings
session = {
    # the cookie key
    'name': 'dev_myGSession',
    # lifetime in seconds. use 0 to have the session last until the
    # until browser is closed or up to 24h
    'lifetime': 60 * 60 * 24 * 14,
    # type (class implementation of session)
    'type': 'express',
}

# cookie settings
cookie = {
    'path': '/',
    # a valid cookie domain must start with a dot
    # and must have at least one embedded dot
    # .localhost is not a valid domain as it does not
    # have an embedded dot
    'domain': 'gengo.dev',
    # only use the cookie over ssl
    'secure': False,
    # use only for http requests, i.e. no javascript
    'httponly': False,
}

# encryption/descryption key, use to encrypt session cookie values
key = 'e77989ed21758e78331b20e477fc5582'

# local timezone
timezone = 'Asia/Tokyo'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base64
import hashlib
from Crypto.Cipher import AES
from Crypto import Random


class Crypt(object):

    """
    Class for simple encryption and decryption using AES cipher in ECB
    Mode
    """

    def __init__(self, key=None):
        self.key = key
        self.padding = "\0"
        self.block_size = 16

    def encrypt(self, data, key=None):
        """ Encrypt the passed data
        data -- the data to encrypt
        key -- (optional if passed from the construct)

        returns the base64 encrypted data
        """
        key = self.get_key(key)
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(key, AES.MODE_ECB, iv)
        return base64.b64encode(iv + cipher.encrypt(self.pad((data))))

    def decrypt(self, data, key=None):
        """ Decrypt the passed data
        data -- the data to decrypt, assumed to be a base64 encrypted string
        key -- (optional if passed from the construct)

        returns the decrypted data
        """
        key = self.get_key(key)
        data = base64.b64decode(data)
        iv = data[0:AES.block_size]
        data = data[AES.block_size:]
        cipher = AES.new(key, AES.MODE_ECB, iv)
        return cipher.decrypt(data).strip(self.padding)

    def pad(self, s):
        return s + (self.block_size - len(s) % self.block_size) * self.padding

    def set_padding(self, padding):
        self.padding = padding

    def set_block_size(self, block_size):
        self.block_size = block_size

    def get_key(self, key=None):
        current_key = key if key else self.key
        if not current_key:
            raise Exception(
                'Crypt class requires a key to perform encryption/decryption operations')
        return self.md5(current_key)

    def md5(self, data):
        m = hashlib.md5()
        m.update(data)
        return m.hexdigest()



import re

from bottle import request
from babel import support, Locale

from .config import Config

# Format of http.request.header.Accept-Language.
# refs: http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.4
REQUEST_ACCEPT_LANGUAGE_RE = re.compile(r'''
        ([A-Za-z]{1,8}(?:-[A-Za-z]{1,8})*|\*)   # "en", "en-au", "*"
        (?:;q=(0(?:\.\d{,3})?|1(?:.0{,3})?))?   # Optional "q=1.00", "q=0.8"
        (?:\s*,\s*|$)                           # Multiple accepts per header.
        ''', re.VERBOSE)


def extra_client_expected_langs():
    """Return language list from http.request.header.Accept-Language,
    ordered by 'q'."""
    result = []
    pieces = REQUEST_ACCEPT_LANGUAGE_RE.split(
        request.headers.get('Accept-Language', ''))
    if pieces[-1]:
        return []
    for i in range(0, len(pieces) - 1, 3):
        first, lang, priority = pieces[i: i + 3]
        if lang == '*':  # TODO: support default language
            return []
        if first:
            return []
        priority = priority and float(priority) or 1.0
        result.append((lang, priority))
    result.sort(key=lambda k: k[1], reverse=True)
    return result


def get_ui_lc():
    """Get language code as defined by domain or session 'lc' variable.
    TODO: generalize and make more flexible.

    Returns None if no suitable setting was found.
    """
    config = Config()
    domain_lc = request.url.split('://')[1].split('.')[0]
    if not domain_lc in config.supported_languages:
        domain_lc = None

    ui_lc = domain_lc

    # if using the express session plugin, override setting with the domain
    # being used: (should be generalized)
    session = request.environ.get('express.session', {})
    if not domain_lc and session.get('lc', None):
        ui_lc = session.get('lc', None)

    return ui_lc


def get_language_list():
    """Return a sorted list of language code-formatted languages in
    descending order of priority.
    """
    config = Config()
    lang_code = getattr(config, 'lang_code', 'en')

    expected_langs = extra_client_expected_langs()

    lang_codes = []

    for lang, priority in expected_langs:
        lang_country = lang.split('-')
        if len(lang_country) == 1:
            lang_codes.append(lang)
            continue
        country = lang_country[1]
        lang_codes.append('%s_%s' % (lang_country[0], country))
        lang_codes.append('%s_%s' % (lang_country[0], country.swapcase()))

    if not lang_code is None:
        lang_codes += [lang_code]

    ui_lc = get_ui_lc()
    if not ui_lc is None:
        lang_codes = [ui_lc] + lang_codes

    return lang_codes


def locale_selector_func():
    lc_list = get_language_list()
    if lc_list:
        return lc_list[0]
    return 'en'


def get_translations():
    """Returns the correct gettext translations that should be used for
    this request.  This will never fail and return a dummy translation
    object if used outside of the request or if a translation cannot be
    found.
    """
    config = Config()
    domain = getattr(config, 'domain', 'locale')
    locale_dir = getattr(config, 'locale_dir')

    translations = request.environ.get('babel_translations', None)
    if translations is None:
        translations = support.Translations.load(locale_dir, [get_locale()], domain=domain)
        request.environ['babel_translations'] = translations
    return translations


def get_locale():
    """Returns the locale that should be used for this request as
    `babel.Locale` object.  This returns `None` if used outside of
    a request.
    """
    config = Config()
    locale = request.environ.get('babel_locale', None)
    if locale is None:
        preferred_lc = locale_selector_func()
        if not preferred_lc:
            locale = getattr(config, 'lang_code', 'en')
        else:
            locale = preferred_lc
        request.environ['babel_locale'] = locale
    return locale


def gettext(string, **variables):
    """Translates a string with the current locale and passes in the
    given keyword arguments as mapping to a string formatting string.

    ::

        gettext(u'Hello World!')
        gettext(u'Hello %(name)s!', name='World')
    """
    t = get_translations()
    if variables:
        if t is None:
            return string % variables
        return t.ugettext(string) % variables
    if t is None:
        return string
    return t.ugettext(string)

_ = gettext


def ngettext(singular, plural, num, **variables):
    """Translates a string with the current locale and passes in the
    given keyword arguments as mapping to a string formatting string.
    The `num` parameter is used to dispatch between singular and various
    plural forms of the message.  It is available in the format string
    as ``%(num)d`` or ``%(num)s``.
    """
    variables.setdefault('num', num)
    t = get_translations()
    if variables:
        if t is None:
            return (singular if num == 1 else plural) % variables
        return t.ungettext(singular, plural, num) % variables
    if t is None:
        return (singular if num == 1 else plural)
    return t.ungettext(singular, plural, num)


def pgettext(context, string, **variables):
    """Like :func:`gettext` but with a context.
    """
    t = get_translations()
    if variables:
        if t is None:
            return string % variables
        return t.upgettext(context, string) % variables
    if t is None:
        return string
    return t.upgettext(context, string)


def npgettext(context, singular, plural, num, **variables):
    """Like :func:`ngettext` but with a context.
    """
    variables.setdefault('num', num)
    t = get_translations()
    if variables:
        if t is None:
            return (singular if num == 1 else plural) % variables
        return t.unpgettext(context, singular, plural, num) % variables
    if t is None:
        return (singular if num == 1 else plural)
    return t.unpgettext(context, singular, plural, num)



from distutils.core import Command
import os
import sys
from setuptools import setup, find_packages
from subprocess import call


class Pep8Command(Command):
    description = "Run pep8 script"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            import pep8
            pep8
        except ImportError:
            print('Missing "pep8" library. You can install it using pip: '
                  'pip install pep8')
            sys.exit(1)

        cwd = os.getcwd()
        retcode = call(("pep8 %s/decanter/ --exclude=vendor --ignore=E501" % (
            cwd)).split(' '))
        sys.exit(retcode)


class TestCommand(Command):
    description = "Run tests"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        errno = subprocess.call(['nosetests', '--debug=DEBUG', '-s', 'tests/unit'])
        raise SystemExit(errno)


class TestCommandWithCoverage(Command):
    description = "Run tests and generate coverage report"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        errno = subprocess.call(['nosetests', '--with-cover', '--cover-html', '--cover-package=decanter', 'tests/unit'])
        raise SystemExit(errno)


# Required repositories
with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as f:
    install_requires = f.read().splitlines()

setup(
    name="decanter",
    version=open('decanter/_version.py').readlines()[-1].split()[-1].strip("\"'"),
    packages=find_packages(),
    scripts=['decanter/decanter-admin.py'],
    install_requires=(
        "Jinja2==2.6",
        "MarkupSafe==0.15",
        "argparse==1.2.1",
        "Babel==1.3",
        "bottle==0.11.6",
        "gevent==0.13.8",
        "greenlet==0.4.0",
        "jsonschema==1.3.0",
        "mock==1.0.1",
        "nose==1.3.0",
        "pep8==1.4.5",
        "phpserialize",
        "pika==0.9.13",
        "pycrypto==2.6",
        "redis==2.7.5",
        "requests==1.2.0",
        "unittest2==0.5.1",
        "wsgiref==0.1.2",
    ),

    package_data={
        # If any package contains *.txt, *.rst or *.md files, include them:
        '': ['*.txt', '*.rst', '*.md'],
    },

    # metadata for upload to PyPI
    author="Gengo",
    author_email="",
    description="A humble web framework based on bottle",
    long_description=open(os.path.join(
        os.path.dirname(__file__), 'README.md')).read(),
    cmdclass={
        'pep8': Pep8Command,
        'test': TestCommand,
        'coverage': TestCommandWithCoverage,
    },
    license="BSD",
    keywords="web framework bottle gengo",
    url="https://github.com/gengo/decanter",   # project home page, if any

    # could also include long_description, download_url, classifiers, etc.
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet'
    ]

)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import uuid
import urllib
import calendar
import phpserialize
from bottle import request
from bottle import response
from .logger import Log
from .crypt import Crypt
from .store import Redis
from .config import Config


class Session(object):

    def __init__(self, session=None):
        self.session = session

    def __getattr__(self, name):
        return getattr(self.session, name)

    def __getitem__(self, key):
        return self.session.get(key, None)

    def __setitem__(self, key, value):
        self.session.set(key, value)

    def __delitem__(self, key):
        self.session.delete(key)

    def __iter__(self):
        return self.session.__iter__()

    def __contains__(self, name):
        return self.session.__contains__(name)

    def __str__(self):
        return self.session.__str__()


class SessionAbstract(object):
    one_day = 60 * 60 * 24

    def __init__(self):
        # initialize logger
        self.log = Log.get_instance()
        # get config
        self.config = Config()

        # set session defaults
        self.name = 'DECANTERSESID'
        self.lifetime = 0
        if 'session' in self.config:
            # name of session cookie
            self.name = self.config.session.get('name', self.name)
            # session lifetime
            self.lifetime = self.config.session.get('lifetime', self.lifetime)

        # set cookie defaults
        self.path = '/'
        self.domain = None
        self.secure = False
        self.httponly = False
        if 'cookie' in self.config:
            # cookie path
            self.path = self.config.cookie.get('path', self.path)
            # cookie domain
            self.domain = self.config.cookie.get('domain', self.domain)
            # cookies secure over ssl only
            self.secure = self.config.cookie.get('secure', self.secure)
            # cookie over http only i.e. no javascript
            self.httponly = self.config.cookie.get('httponly', self.httponly)

        # the session data
        self.data = {}

    def write(self):
        raise NotImplementedError("write method not implemented")

    def read(self):
        raise NotImplementedError("read method not implemented")

    def get(self, key, default=None):
        return self.data.get(key, default)

    def get_once(self, key, default=None):
        value = self.data.get(key, default)
        self.delete(key)
        return value

    def has(self, key):
        return key in self.data

    def set(self, key, value):
        self.data[key] = value

    def delete(self, key):
        try:
            del(self.data[key])
            return True
        except:
            return False

    def clear(self):
        self.data = {}

    def __getitem__(self, key):
        return self.data.get(key, None)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        try:
            del(self.data[key])
        except:
            pass

    def __iter__(self):
        return self.data.__iter__()

    def __contains__(self, name):
        return name in self.data

    def __str__(self):
        return self.data.__str__()


class ExpressSession(SessionAbstract):

    def __init__(self, cookies):
        super(ExpressSession, self).__init__()
        # SimpleCookie object
        self.cookies = cookies
        # initialize crypt library
        self.crypt = Crypt(self.config.key)
        # redis session storage
        self.redis = Redis()
        # the encrypted cookie value
        self.cookie = {'session_id': None,
                       'ip_address': None,
                       'user_agent': None,
                       'last_activity': None}

    def read(self):
        cookie = False
        if self.name in self.cookies:
            cookie = self.cookies[self.name].value
        try:
            if cookie:
                cookie = urllib.unquote(cookie)
                cookie = self.crypt.decrypt(cookie)
                self.cookie = phpserialize.loads(cookie)
                if 'session_id' in self.cookie:
                    data = self.redis.get(self.cookie['session_id'])
                    if data:
                        self.data = phpserialize.loads(
                            data, object_hook=phpserialize.phpobject)
                    else:
                        self.data = {}
                        for key in self.cookie:
                            self.cookie[key] = None

        except Exception as e:
            print("Error while reading session: {0}".format(e))
            self.log.error("Error while reading session: {0}".format(e))

    def write(self):
        if self.cookie['session_id']:
            self.save()
        else:
            self.create()

    def save(self):
        timestamp = calendar.timegm(time.gmtime())
        max_age = self.lifetime if self.lifetime else self.one_day
        expires = timestamp + max_age

        params = {}
        if max_age:
            params['max_age'] = max_age
        if expires:
            params['expires'] = expires
        params['domain'] = self.domain
        params['path'] = self.path
        if self.secure:
            params['secure'] = self.secure
        if self.httponly:
            params['httponly'] = self.httponly

        data = phpserialize.dumps(
            self.data, object_hook=phpserialize.phpobject)
        self.redis.set(self.cookie['session_id'], data)
        response.set_cookie(self.name, urllib.quote_plus(
            self.crypt.encrypt(phpserialize.dumps(self.cookie))), **params)

    def create(self):
        timestamp = calendar.timegm(time.gmtime())

        self.cookie['session_id'] = self.crypt.md5(uuid.uuid4().get_bytes())
        self.cookie['ip_address'] = request.remote_addr
        self.cookie['user_agent'] = request.environ.get('HTTP_USER_AGENT', '')[0:50]
        self.cookie['last_activity'] = timestamp

        max_age = self.lifetime if self.lifetime else self.one_day
        expires = timestamp + max_age

        params = {}
        if max_age:
            params['max_age'] = max_age
        if expires:
            params['expires'] = expires
        params['domain'] = self.domain
        params['path'] = self.path
        if self.secure:
            params['secure'] = self.secure
        if self.httponly:
            params['httponly'] = self.httponly

        data = phpserialize.dumps(
            self.data, object_hook=phpserialize.phpobject)
        self.redis.set(self.cookie['session_id'], data)
        self.redis.expire(self.cookie['session_id'], max_age)
        response.set_cookie(self.name, urllib.quote_plus(
            self.crypt.encrypt(phpserialize.dumps(self.cookie))), **params)

    def __str__(self):
        return ':'.join([self.cookie['session_id'] if self.cookie['session_id'] else '', self.data.__str__()])
